from enum import Enum, IntFlag, auto


class CharacterType(Enum):
    """
    Represents the types of characters recognized during parsing.

    #### Members
    - `NUMBER` — Represents number characters.
    - `ALPHABET` — Represents alphabetic characters.
    - `SYMBOL` — Represents symbol characters.
    """

    NUMBER = "NUMBER"
    ALPHABET = "ALPHABET"
    SYMBOL = "SYMBOL"


class ValueType(Enum):
    """
    Represents the types of strings recognized during parsing.

    #### Members
    - `NUMBER` — Represents number strings (ex: 123).
    - `NUMERAL` — Represents numeral strings (ex: twenty).
    - `SCALE` — Represents scale strings (ex: minutes).
    - `SYMBOL` — Represents symbol strings (ex: !$^).
    - `MIXED` — Represents mixed strings.
    """

    NUMBER = "NUMBER"
    NUMERAL = "NUMERAL"
    SCALE = "SCALE"
    SYMBOL = "SYMBOL"
    MIXED = "MIXED"


class NumeralType(Enum):
    """
    Represents the types of numerals recognized during parsing.

    #### Members
    - `DIGIT` — Represents digit numerals (ex: one).
    - `TEEN` — Represents teen numerals (ex: eleven).
    - `TEN` — Represents ten numerals (ex: twenty).
    - `HUNDRED` — Represents hundred numerals (ex: hundred).
    - `THOUSAND` — Represents thousand numerals (ex: thousand).
    - `MULTIPLIER` — Represents multiplier numerals (ex: half).
    - `OPERATOR` — Represents operator numerals (ex: of).
    - `NONE` — Represents no numeral type.
    """

    DIGIT = "DIGIT"
    TEEN = "TEEN"
    TEN = "TEN"
    HUNDRED = "HUNDRED"
    THOUSAND = "THOUSAND"
    MULTIPLIER = "MULTIPLIER"
    OPERATOR = "OPERATOR"
    NONE = "NONE"


class FailureFlags(IntFlag):
    """
    Represents the inputs which will cause parsing to fail.

    #### Members
    - `NONE` — No failures will cause parsing to fail.
    - `ALL` — Any failure will cause parsing to fail.
    - `MALFORMED_CONTENT` — The fallback when something that shouldn't have happened, happened.
    - `UNKNOWN_TERM` — The parsed value was not recognized from the terms or symbols in the config.
    - `MALFORMED_DECIMAL` — Multiple decimals were attempted within a singular decimal segment.
    - `MALFORMED_THOUSAND` — A thousand segment was attempted but did not have a leading number or a proper number of digits following a thousand separator.
    - `MALFORMED_FRACTION` — A fraction was attempted but had more than 2 values, a missing value, or was not formatted correctly.
    - `MALFORMED_HHMMSS` — An HH:MM:SS was attempted but had more segments than enabled scales or was not formatted correctly.
    - `LONELY_VALUE` — A value was parsed with no paired scale.
    - `LONELY_SCALE` — A scale was parsed with no paired value.
    - `LEADING_SCALE` — A scale was parsed at the beginning of the input.
    - `DUPLICATE_SCALE` — The same scale was parsed multiple times.
    - `CONSECUTIVE_CONNECTOR` — More than the allowed number of connectors were parsed in a row.
    - `CONSECUTIVE_SEGMENTORS` — More than the allowed number of segmentors were parsed in a row.
    - `CONSECUTIVE_SPECIALS` — More than the allowed number of special characters were parsed in a row.
    - `MISPLACED_ALLOWED_TERM` — An allowed term was found in the middle of a segment/sentence.
    - `MISPLACED_SPECIAL` — A special character was found in the middle of a segment/sentence.
    - `UNUSED_OPERATOR` — A term of the operator numeral was parsed but unused on any values.
    - `AMBIGUOUS_MULTIPLIER` — More than one multiplier was parsed for a single segment which may be ambiguous.
    """

    NONE = 0
    MALFORMED_CONTENT = auto()
    UNKNOWN_TERM = auto()
    MALFORMED_DECIMAL = auto()
    MALFORMED_THOUSAND = auto()
    MALFORMED_FRACTION = auto()
    MALFORMED_HHMMSS = auto()
    LONELY_VALUE = auto()
    LONELY_SCALE = auto()
    DUPLICATE_SCALE = auto()
    CONSECUTIVE_CONNECTOR = auto()
    CONSECUTIVE_SEGMENTOR = auto()
    CONSECUTIVE_SPECIAL = auto()
    MISPLACED_ALLOWED_TERM = auto()
    MISPLACED_SPECIAL = auto()
    UNUSED_OPERATOR = auto()
    AMBIGUOUS_MULTIPLIER = auto()
    ALL = (
        MALFORMED_CONTENT
        | UNKNOWN_TERM
        | MALFORMED_DECIMAL
        | MALFORMED_THOUSAND
        | MALFORMED_FRACTION
        | MALFORMED_HHMMSS
        | LONELY_VALUE
        | LONELY_SCALE
        | DUPLICATE_SCALE
        | CONSECUTIVE_CONNECTOR
        | CONSECUTIVE_SEGMENTOR
        | CONSECUTIVE_SPECIAL
        | MISPLACED_ALLOWED_TERM
        | MISPLACED_SPECIAL
        | UNUSED_OPERATOR
        | AMBIGUOUS_MULTIPLIER
    )

    def _get_flags(self) -> list:
        return [f"FailureFlags.{flag._name_}" for flag in FailureFlags if flag and flag in self]

    def __str__(self) -> str:
        return " | ".join(self._get_flags()).replace("FailureFlags.", "") or "NONE"

    def __repr__(self) -> str:
        return "(" + (" | ".join(self._get_flags()) or "FailureFlags.NONE") + ")"



from __future__ import annotations

import json
from dataclasses import dataclass
from datetime import timedelta
from typing import Literal

from timelength.enums import FailureFlags, NumeralType


@dataclass
class Scale:
    """---
    Represents a time scale.

    #### Attributes
    - scale: `float = 0.0` — The number of seconds in the scale.
    - singular: `str = ""` — The singular form of the scale.
    - plural: `str = ""` — The plural form of the scale.
    - terms: `tuple[str, ...] = ()` — A tuple of terms associated with the scale.
    - enabled: `bool = True` — Whether the scale is enabled or not.

    #### Properties
    - valid: `bool` — Return whether the scale is valid or not based on all attributes being set.
    """

    scale: float = 0.0
    singular: str = ""
    plural: str = ""
    terms: tuple[str, ...] = ()
    enabled: bool = True

    @property
    def valid(self) -> bool:
        """Return whether the scale is valid or not based on all attributes being set."""
        return bool(self.scale and self.singular and self.plural and self.terms)

    def __str__(self) -> str:
        """Return the title cased singular form of the scale."""
        return self.singular.title()

    def __repr__(self) -> str:
        """Return a string representation of the scale with attributes included."""

        def tuple_to_str(tup: tuple) -> str:
            return "(" + ", ".join([json.dumps(term) for term in tup]) + ")"

        terms_str = tuple_to_str(self.terms)
        return f"Scale(scale={self.scale}, singular={json.dumps(self.singular)}, plural={json.dumps(self.plural)}, terms={terms_str}, enabled={self.enabled})"

    def __eq__(self, value: object) -> bool:
        """Return whether the scale is equal to the value."""
        if not isinstance(value, Scale):
            return False
        return self.scale == value.scale

    def __hash__(self) -> int:
        """Return the hash of the scale."""
        return hash((self.scale, self.singular, self.plural, self.terms, self.enabled))


@dataclass
class Numeral:
    """---
    Represents a number related word.

    #### Attributes
    - name: `str = ""` — The name of the numeral.
    - type: `NumeralType = NumeralType.NONE` — The type of numeral.
    - value: `float = 0.0` — The value of the numeral.
    - terms: `tuple[str, ...] = []` — A list of terms associated with the numeral.

    #### Properties
    - valid: `bool` — Return whether the numeral is valid or not based on all attributes being set.
    """

    name: str = ""
    type: NumeralType = NumeralType.NONE
    value: float = 0.0
    terms: tuple[str, ...] = ()
    enabled: bool = True

    @property
    def valid(self) -> bool:
        """Return whether the numeral is valid or not based on all attributes being set."""
        return bool(
            self.name and self.type and self.type is not NumeralType.NONE and self.value is not None and self.terms
        )

    def __str__(self) -> str:
        """Return the title cased singular form of the numeral."""
        return self.name.title()

    def __repr__(self) -> str:
        """Return a string representation of the numeral with attributes included."""

        def tuple_to_str(tup: tuple) -> str:
            return "(" + ", ".join([json.dumps(term) for term in tup]) + ")"

        terms_str = tuple_to_str(self.terms)
        return f"Numeral(name={json.dumps(self.name)}, type={self.type}, value={self.value}, terms={terms_str}, enabled={self.enabled})"

    def __eq__(self, value: object) -> bool:
        """Return whether the numeral is equal to the value."""
        if not isinstance(value, Numeral):
            return False
        return self.value == value.value

    def __hash__(self) -> int:
        """Return the hash of the numeral."""
        return hash((self.name, self.type, self.value, self.terms, self.enabled))


class ParsedTimeLength:
    """
    Represents the outcome of parsing a TimeLength.

    #### Properties
    - success: `str` — Indicates if the parsing was successful.
    - seconds: `float` — The total length of time parsed, in seconds.
    - delta: `timedelta | None` — The parsed time as a `timedelta`, or `None` if out of range.
    - invalid: `tuple` — The parts of the input string that could not be parsed as valid time.
    - valid: `tuple` — The parts of the input string that were successfully parsed as valid time.
    """

    def __init__(
        self,
        success: bool = False,
        seconds: float = 0.0,
        delta: timedelta | None = None,
        invalid: tuple[tuple[float | str, FailureFlags], ...] = (),
        valid: tuple[tuple[float, Scale], ...] = (),
    ) -> None:
        self._success: bool = success
        self._seconds: float = seconds
        self._delta: timedelta | None = delta
        self._invalid: tuple[tuple[float | str, FailureFlags], ...] = invalid
        self._valid: tuple[tuple[float, Scale], ...] = valid

    @property
    def success(self) -> bool:
        """Return if the parsing was successful."""
        return self._success

    @property
    def seconds(self) -> float:
        """Return the total length of time parsed, in seconds."""
        return self._seconds

    @property
    def delta(self) -> timedelta | None:
        """Return a timedelta representing the parsed time, or None if out of range."""
        return self._delta

    @property
    def invalid(self) -> tuple[tuple[float | str, FailureFlags], ...]:
        """Return a tuple of parts of the input string that could not be parsed as valid time."""
        return self._invalid

    @property
    def valid(self) -> tuple[tuple[float, Scale], ...]:
        """Return a tuple of parts of the input string that were successfully parsed as valid time."""
        return self._valid

    def __str__(self) -> str:
        """Return a string indicating the success or failure of the parsing."""
        return "Success" if self.success else "Failure"

    def __repr__(self) -> str:
        """Return a string representation of the ParsedTimeLength with attributes included."""

        def tuple_to_str(tup: tuple) -> str:
            return (
                "("
                + ", ".join(
                    [
                        tuple_to_str(term)
                        if isinstance(term, tuple)
                        else repr(term)
                        if isinstance(term, Scale)
                        else json.dumps(term)
                        for term in tup
                    ]
                )
                + ")"
            )

        valid_str = tuple_to_str(self.valid)
        invalid_str = tuple_to_str(self.invalid)
        return f"ParsedTimeLength(success={self.success}, seconds={self.seconds}, invalid={invalid_str}, valid={valid_str})"


@dataclass
class Buffer:
    """---
    Represents the state of the buffer during parsing.

    #### Attributes
    - value: `str = ""` — The current value of the buffer.
    """

    value: str = ""

    def __str__(self) -> str:
        """Return the current value of the buffer."""
        return self.value

    def __repr__(self) -> str:
        """Return a string representation of the buffer with attributes included."""
        return f'Buffer(value="{self.value}")'


@dataclass
class ParserSettings:
    """---
    Represents the settings for the parser.

    #### Attributes
    - assume_scale: `Literal["LAST", "SINGLE", "NEVER"] = "SINGLE"` — How to handle no scale being provided.
        - `LAST` will assume seconds only for the last value if no scale is provided for it.
        - `SINGLE` will only assume seconds when a single input is provided.
        - `NEVER` will never assume seconds when no scale is provided.
    - limit_allowed_terms: `bool = True` — Prevent terms from the `allowed_terms` list in
        the config from being used in the middle of a segment, thus interrupting a value/scale pair.
        - The affected segment will become abandoned and added to the invalid list.
        - The terms may still be used at the beginning or end of a segment/sentence.
        - If `False`, The terms will be ignored (within other limitations) and not effect parsing.
    - allow_duplicate_scales: `bool = True` — Allow scales to be parsed multiple times, stacking their values.
        - If `False`, the first scale will be used and subsequent duplicates will be added to the invalid list.
    - allow_thousands_extra_digits: `bool = False` — Allow thousands to be parsed with more than three digits
        following the thousand delimiter.
        - Ex: `1,2345` will be interpreted as `12,345`.
    - allow_thousands_lacking_digits: `bool = False` — Allow thousands to be parsed with less than three digits.
        - Ex: `1,23` will be interpreted as `123`.
    - allow_decimals_lacking_digits: `bool = True` — Allow decimals to be parsed with no number following the
        decimal delimiter.
        - Ex: `1.` will be interpreted as `1.0`.
    """

    assume_scale: Literal["LAST", "SINGLE", "NEVER"] = "SINGLE"
    limit_allowed_terms: bool = True
    allow_duplicate_scales: bool = True
    allow_thousands_extra_digits: bool = False
    allow_thousands_lacking_digits: bool = False
    allow_decimals_lacking_digits: bool = True

    def __str__(self) -> str:
        """Return the settings as either 'Default' or 'Modified' based on the default values."""
        return (
            "Modified"
            if any(
                [
                    self.assume_scale != "SINGLE",
                    self.limit_allowed_terms is False,
                    self.allow_duplicate_scales is False,
                    self.allow_thousands_extra_digits is True,
                    self.allow_thousands_lacking_digits is True,
                    self.allow_decimals_lacking_digits is False,
                ]
            )
            else "Default"
        )

    def __repr__(self) -> str:
        """Return a string representation of the ParserSettings with attributes included."""
        return (
            f'ParserSettings(assume_scale="{self.assume_scale}", '
            + f"limit_allowed_terms={self.limit_allowed_terms}, "
            + f"allow_duplicate_scales={self.allow_duplicate_scales}, "
            + f"allow_thousands_extra_digits={self.allow_thousands_extra_digits}, "
            + f"allow_thousands_lacking_digits={self.allow_thousands_lacking_digits}, "
            + f"allow_decimals_lacking_digits={self.allow_decimals_lacking_digits})"
        )



"""A flexible python duration parser designed for human readable lengths of time."""

from timelength.dataclasses import Numeral, ParsedTimeLength, ParserSettings, Scale
from timelength.enums import FailureFlags
from timelength.errors import (
    InvalidLocaleError,
    InvalidNumeralError,
    InvalidParserError,
    InvalidScaleError,
    NotALocaleError,
    NoValidScalesError,
    ParsedTimeDeltaError,
    PotentialDateTimeError,
    PotentialTimeDeltaError,
)
from timelength.locales import English, Guess, Locale, Spanish
from timelength.timelength import TimeLength



from __future__ import annotations

import json
import os
from fractions import Fraction
from importlib import util
from typing import Any, Callable

from timelength.dataclasses import Numeral, ParsedTimeLength, ParserSettings, Scale
from timelength.enums import FailureFlags, NumeralType
from timelength.errors import (
    InvalidLocaleError,
    InvalidNumeralError,
    InvalidParserError,
    InvalidScaleError,
    NoValidScalesError,
)


class Locale:
    """---
    Represents a `Locale` context used for parsing lengths of time.

    #### Attributes
    - config_location: `str | os.PathLike = "english.json"` — The path to the config file for this `Locale`.
    - flags: `FailureFlags | None = None` — The flags that will cause parsing to fail.
        - If passed, flags loaded from the config will be overwritten.
    - settings: `ParserSettings | None = None` — The settings for the parser.
        - If passed, settings loaded from the config will be overwritten.

    #### Properties
    - `parser` — Return the parser function attached to the Locale.
    - `connectors` — Return the connectors if valid.
    - `segmentors` — Return the segmentors if valid.
    - `allowed_terms` — Return the allowed terms.
    - `hhmmss_delimiters` — Return the hhmmss delimiters if valid.
    - `decimal_delimiters` — Return the decimal delimiters if valid.
    - `thousand_delimiters` — Return the thousand delimiters if valid.
    - `fraction_delimiters` — Return the fraction delimiters if valid.
    - `delimiters` — Return a list of all delimiters if valid.
    - `specials` — Return a list of all special characters.
    - `usable_scales` — Return a list of valid and enabled scales.
    - `usable_numerals` — Return a list of valid and enabled numerals.
    - `base_scale` — Return the base scale for the locale.

    #### Methods
    - `get_scale()` — Get the scale that contains a specific value in its terms list, if any.
    - `get_numeral()` — Get the numeral that contains a specific value in its terms list, if any.

    #### Raises
    - `InvalidLocaleError` when the config at `config_location` does not exist or is malformed.
    - `InvalidParserError` when `self.parser` is not callable.
    - `NoValidScalesError` when no `Scale` is valid during validation.
    - `InvalidScaleError` when a `Scale` is invalid during initialization.
    - `InvalidNumeralError` when a `Numeral` is invalid during initialization.
    """

    def __init__(
        self,
        config_location: str | os.PathLike[str] = "english.json",
        flags: FailureFlags | None = None,
        settings: ParserSettings | None = None,
    ) -> None:
        self._config_location: str = (
            os.fspath(config_location) if isinstance(config_location, os.PathLike) else config_location
        )
        base_dir: str = os.path.dirname(__file__)
        locale_path: str = os.path.join(base_dir, "locales", self._config_location)
        full_json_path: str = locale_path if os.path.exists(locale_path) else self._config_location
        self._config: dict[str, Any] = self._load_config(full_json_path)

        self._parser_file: str = self._get_config_or_raise("parser_file", str)
        self._parser: Callable[[str, Locale], ParsedTimeLength] = self._load_parser(base_dir)

        self.flags: FailureFlags = self._load_flags(flags)
        self.settings: ParserSettings = self._load_settings(settings)

        self._connectors: list[str] = self._get_config_or_raise("connectors", list)
        self._segmentors: list[str] = self._get_config_or_raise("segmentors", list)
        self._validate_connectors_segmentors()

        self._allowed_terms: list[str] = self._get_config_or_raise("allowed_terms", list)
        self._hhmmss_delimiters: list[str] = self._get_config_or_raise("hhmmss_delimiters", list)
        self._decimal_delimiters: list[str] = self._get_config_or_raise("decimal_delimiters", list)
        self._thousand_delimiters: list[str] = self._get_config_or_raise("thousand_delimiters", list)
        self._fraction_delimiters: list[str] = self._get_config_or_raise("fraction_delimiters", list)
        self._validate_delimiters()

        self._scales_json = self._get_config_or_raise("scales", dict)
        self.millisecond: Scale
        self.second: Scale
        self.minute: Scale
        self.hour: Scale
        self.day: Scale
        self.week: Scale
        self.month: Scale
        self.year: Scale
        self.decade: Scale
        self.century: Scale
        self.scales: list[Scale] = []
        self._load_scales(self._scales_json)
        self._validate_scales()

        self._numerals_json: dict[str, dict[str, Any]] = self._get_config_or_raise("numerals", dict)
        self.numerals: list[Numeral] = []
        self._load_numerals(self._numerals_json)

        self.extra_data: dict[str, Any] = self._get_config_or_raise("extra_data", dict)

    @property
    def parser(self) -> Callable[[str, Locale], ParsedTimeLength]:
        """Return the parser function attached to the Locale."""

        if self._parser and callable(self._parser):
            self._validate()
            return self._parser

        raise InvalidParserError(self)

    @property
    def connectors(self) -> list[str]:
        """Return the connectors if valid."""
        self._validate_connectors_segmentors()
        return self._connectors

    @connectors.setter
    def connectors(self, value: list[str]) -> None:
        self._connectors = value
        self._validate_connectors_segmentors()

    @property
    def segmentors(self) -> list[str]:
        """Return the segmentors if valid."""
        self._validate_connectors_segmentors()
        return self._segmentors

    @segmentors.setter
    def segmentors(self, value: list[str]) -> None:
        self._segmentors = value
        self._validate_connectors_segmentors()

    @property
    def allowed_terms(self) -> list[str]:
        """Return the allowed terms."""
        return self._allowed_terms

    @allowed_terms.setter
    def allowed_terms(self, value: list[str]) -> None:
        self._allowed_terms = value

    @property
    def decimal_delimiters(self) -> list[str]:
        """Return the decimal delimiters if valid."""
        self._validate_delimiters()
        return self._decimal_delimiters

    @decimal_delimiters.setter
    def decimal_delimiters(self, value: list[str]) -> None:
        self._decimal_delimiters = value
        self._validate_delimiters()

    @property
    def thousand_delimiters(self) -> list[str]:
        """Return the thousand delimiters if valid."""
        self._validate_delimiters()
        return self._thousand_delimiters

    @thousand_delimiters.setter
    def thousand_delimiters(self, value: list[str]) -> None:
        self._thousand_delimiters = value
        self._validate_delimiters()

    @property
    def hhmmss_delimiters(self) -> list[str]:
        """Return the hhmmss delimiters if valid."""
        self._validate_delimiters()
        return self._hhmmss_delimiters

    @hhmmss_delimiters.setter
    def hhmmss_delimiters(self, value: list[str]) -> None:
        self._hhmmss_delimiters = value
        self._validate_delimiters()

    @property
    def fraction_delimiters(self) -> list[str]:
        """Return the fraction delimiters if valid."""
        self._validate_delimiters()
        return self._fraction_delimiters

    @fraction_delimiters.setter
    def fraction_delimiters(self, value: list[str]) -> None:
        self._fraction_delimiters = value
        self._validate_delimiters()

    @property
    def delimiters(self) -> list[str]:
        """Return a list of all delimiters if valid."""
        self._validate_delimiters()
        return (
            self._hhmmss_delimiters + self._decimal_delimiters + self._thousand_delimiters + self._fraction_delimiters
        )

    @property
    def specials(self) -> list[str]:
        """Return a list of all special characters."""
        return self._connectors + self._segmentors + self._allowed_terms + self.delimiters

    @property
    def usable_scales(self) -> list[Scale]:
        """Return a list of usable scales."""
        return [scale for scale in self.scales if scale.valid and scale.enabled]

    @property
    def usable_numerals(self) -> list[Numeral]:
        """Return a list of usable numerals."""
        return [numeral for numeral in self.numerals if numeral.valid and numeral.enabled]

    @property
    def base_scale(self) -> Scale:
        """---
        Return the base `Scale` for `self`.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found.

        #### Returns
        - The base `Scale` for the locale. By default this is `Second`, but if it is not valid or enabled,
        the first valid and enabled scale will be used.
        """

        base = self.second if self.second.enabled and self.second.valid else None

        if not base:
            usable_scales = self.usable_scales
            base = usable_scales[0] if usable_scales else None

            if not base:
                raise NoValidScalesError

        return base

    def get_scale(self, term: str) -> Scale | None:
        """---
        Get a `Scale` from a term.

        #### Raises
        - `InvalidScaleError` when the scale is not valid.

        #### Returns
        - The `Scale` that contains the term in its terms list.
        - `None` if no scale contains the term.
        """

        for scale in self.scales:
            if term in scale.terms:
                if not scale.valid:
                    raise InvalidScaleError(scale.singular)
                return scale

        return None

    def get_numeral(self, term: str) -> Numeral | None:
        """---
        Get a `Numeral` from a term.

        #### Raises
        - `InvalidNumeralError` when the numeral is not valid.

        #### Returns
        - The `Numeral` that contains the term in its terms list.
        - `None` if no numeral contains the term.
        """

        for numeral in self.numerals:
            if term in numeral.terms:
                if not numeral.valid:
                    raise InvalidNumeralError(numeral.name)
                return numeral

        return None

    def _load_config(self, json_path: str) -> dict[str, Any]:
        if not os.path.exists(json_path) or not os.path.isfile(json_path):
            raise InvalidLocaleError(self, f"The provided config does not exist: {repr(json_path)}")

        with open(json_path, "r", encoding="utf-8") as f:
            config: dict[str, Any] = json.load(f)

        if not config:
            raise InvalidLocaleError(self, f"The provided config is empty: {repr(json_path)}")

        return config

    def _load_parser(self, base_dir: str) -> Callable[[str, Locale], ParsedTimeLength]:
        """Load the parser function from the parser file."""

        parser_path = os.path.join(base_dir, "parsers", self._parser_file)
        full_parser_path: str | None = (
            parser_path
            if os.path.exists(parser_path)
            else self._parser_file
            if os.path.exists(self._parser_file)
            else None
        )
        module_name, _ = os.path.splitext(os.path.basename(full_parser_path)) if full_parser_path else ("", "")

        try:
            if not full_parser_path:
                raise FileNotFoundError

            spec = util.spec_from_file_location(module_name, full_parser_path)

            if not spec or not hasattr(spec, "loader") or not spec.loader:
                raise FileNotFoundError

            module = util.module_from_spec(spec)
            spec.loader.exec_module(module)
            parser: Callable[[str, Locale], ParsedTimeLength] | None = getattr(module, module_name, None)

            if not parser:
                raise AttributeError
            else:
                return parser
        except (ModuleNotFoundError, FileNotFoundError) as e:
            raise InvalidLocaleError(self, f"Parser file not found: {repr(self._parser_file)}") from e
        except AttributeError as e:
            raise InvalidLocaleError(
                self, f"Parser function {repr(module_name)} not found in parser file: {repr(self._parser_file)}"
            ) from e

    def _load_flags(self, flags: FailureFlags | None) -> FailureFlags:
        """Load the flags from the config or the passed flags."""

        if flags is None:
            combined_flags = FailureFlags(0)

            for flag in self._get_config_or_raise("flags", list):
                try:
                    combined_flags |= FailureFlags[flag]
                except KeyError as e:
                    raise InvalidLocaleError(self, f"Invalid flag provided in config: {repr(flag)}") from e

            return combined_flags

        return flags

    def _load_settings(self, settings: ParserSettings | None) -> ParserSettings:
        """Load the settings from the config or the passed settings."""

        if settings is None:
            try:
                settings = ParserSettings(**self._get_config_or_raise("settings", dict))
            except TypeError as e:
                if "got an unexpected keyword argument" in str(e):
                    key: str = str(e).split("'")[1]
                    raise InvalidLocaleError(self, f"Invalid setting provided in config: {repr(key)}") from e
                else:
                    raise InvalidLocaleError(self, "Invalid settings provided in config.") from e

        return settings

    def _load_scales(self, scales_json: dict) -> None:
        """Load the scales from the config."""

        def process_scale(scale_json: dict, scale_name: str) -> None:
            if not scale_name:
                raise InvalidScaleError(scale_name, "A scale key is empty.")
            elif not all(key in scale_json for key in ["scale", "singular", "plural", "terms"]):
                raise InvalidScaleError(scale_name)
            else:
                try:
                    if isinstance(scale_json["scale"], str):
                        scale_json["scale"] = float(Fraction(scale_json["scale"]))

                    scale_json["terms"] = tuple(scale_json["terms"])
                    scale_obj: Scale = Scale(**scale_json)

                    setattr(self, scale_name, scale_obj)
                    self.scales.append(scale_obj)
                except (ValueError, TypeError) as e:
                    raise InvalidScaleError(scale_name, f"{repr(scale_name)} contains invalid values.") from e

        default_scales = [
            "millisecond",
            "second",
            "minute",
            "hour",
            "day",
            "week",
            "month",
            "year",
            "decade",
            "century",
        ]

        if any(scale_name not in scales_json for scale_name in default_scales):
            raise InvalidLocaleError(self, "Default scales are missing from the config.")

        for scale_name in default_scales:
            scale_json = scales_json.get(scale_name, {})
            process_scale(scale_json, scale_name)

        for scale_name in scales_json:
            if scale_name not in default_scales:
                process_scale(scales_json[scale_name], scale_name)

    def _load_numerals(self, numerals_json: dict) -> None:
        """Load the numerals from the config."""

        for numeral_name in numerals_json:
            if not numeral_name:
                raise InvalidNumeralError(numeral_name, "A numeral key is empty.")
            elif not all(key in numerals_json[numeral_name] for key in ["type", "value", "terms"]):
                raise InvalidNumeralError(numeral_name)

            raw_type: str = numerals_json[numeral_name]["type"]

            try:
                if not raw_type or not isinstance(raw_type, str):
                    raise ValueError

                numeral_type: NumeralType = NumeralType(raw_type)
            except ValueError as e:
                raise InvalidNumeralError(
                    numeral_name, f"An invalid NumeralType was provided for {repr(numeral_name)}."
                ) from e

            value: float = (
                float(Fraction(numerals_json[numeral_name]["value"]))
                if isinstance(numerals_json[numeral_name]["value"], str)
                else float(numerals_json[numeral_name]["value"])
            )
            terms: tuple[str, ...] = tuple(numerals_json[numeral_name]["terms"])
            self.numerals.append(Numeral(numeral_name, numeral_type, value, terms))

    def _get_config_or_raise(self, key: str, target_type: type) -> Any:
        """Retrieve a value of specified type from the config or raise `InvalidLocaleError` if the value is invalid."""

        value = self._config.get(key)

        if value is None:
            raise InvalidLocaleError(self, f"No {repr(key)} key found in config.")

        if not isinstance(value, target_type):
            raise InvalidLocaleError(self, f"{repr(key)} key in config is not of type {repr(target_type.__name__)}.")

        return value

    def _validate_connectors_segmentors(self) -> bool:
        """Validate the values for the connectors and segmentors."""

        if set(self._connectors).intersection(self._segmentors):
            raise InvalidLocaleError(self, "Connectors and segmentors may not have overlap.")

        return True

    def _validate_delimiters(self) -> bool:
        """Validate the values for the delimiters."""

        all_delimiters = (
            self._decimal_delimiters + self._thousand_delimiters + self._hhmmss_delimiters + self._fraction_delimiters
        )

        if len(all_delimiters) != len(set(all_delimiters)):
            raise InvalidLocaleError(self, "Delimiters may not have overlap.")

        return True

    def _validate_scales(self) -> bool:
        """Validate the values for the scales."""

        if not self.usable_scales:
            raise NoValidScalesError

        return True

    def _validate(self) -> bool:
        """Validate the values for config attributes."""

        self._validate_connectors_segmentors()
        self._validate_delimiters()
        self._validate_scales()

        return True

    def __str__(self) -> str:
        """Return the name of the locale."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        """Return a string representation of the locale with attributes included."""
        return f"{self.__str__()}(config_location={json.dumps(self._config_location)}, flags={repr(self.flags)}, settings={repr(self.settings)})"


class English(Locale):
    """---
    Represents the `English` context used for parsing lengths of time.

    #### Attributes
    - flags: `FailureFlags | None = None` — The flags that will cause parsing to fail.
        - If passed, flags loaded from the config will be overwritten.
    - settings: `ParserSettings | None = None` — The settings for the parser.
        - If passed, settings loaded from the config will be overwritten.

    #### Properties
    - `parser` — Return the parser function attached to the Locale.
    - `connectors` — Return the connectors if valid.
    - `segmentors` — Return the segmentors if valid.
    - `allowed_terms` — Return the allowed terms.
    - `hhmmss_delimiters` — Return the hhmmss delimiters if valid.
    - `decimal_delimiters` — Return the decimal delimiters if valid.
    - `thousand_delimiters` — Return the thousand delimiters if valid.
    - `fraction_delimiters` — Return the fraction delimiters if valid.
    - `delimiters` — Return a list of all delimiters if valid.
    - `specials` — Return a list of all special characters.
    - `usable_scales` — Return a list of valid and enabled scales.
    - `usable_numerals` — Return a list of valid and enabled numerals.
    - `base_scale` — Return the base scale for the locale.

    #### Methods
    - `get_scale()` — Get the scale that contains a specific value in its terms list, if any.
    - `get_numeral()` — Get the numeral that contains a specific value in its terms list, if any.

    #### Raises
    - `InvalidLocaleError` when `english.json` does not exist or is malformed.
    - `InvalidParserError` when `self.parser` is not callable.
    - `NoValidScalesError` when no `Scale` is valid during validation.
    - `InvalidScaleError` when a `Scale` is invalid during initialization.
    - `InvalidNumeralError` when a `Numeral` is invalid during initialization.
    """

    def __init__(self, flags: FailureFlags | None = None, settings: ParserSettings | None = None) -> None:
        super().__init__("english.json", flags, settings)

    def __repr__(self) -> str:
        """Return a string representation of the locale with attributes included."""
        return f"English(flags={repr(self.flags)}, settings={repr(self.settings)})"


class Spanish(Locale):
    """---
    Represents the `Spanish` context used for parsing lengths of time.

    #### Attributes
    - flags: `FailureFlags | None = None` — The flags that will cause parsing to fail.
        - If passed, flags loaded from the config will be overwritten.
    - settings: `ParserSettings | None = None` — The settings for the parser.
        - If passed, settings loaded from the config will be overwritten.

    #### Properties
    - `parser` — Return the parser function attached to the Locale.
    - `connectors` — Return the connectors if valid.
    - `segmentors` — Return the segmentors if valid.
    - `allowed_terms` — Return the allowed terms.
    - `hhmmss_delimiters` — Return the hhmmss delimiters if valid.
    - `decimal_delimiters` — Return the decimal delimiters if valid.
    - `thousand_delimiters` — Return the thousand delimiters if valid.
    - `fraction_delimiters` — Return the fraction delimiters if valid.
    - `delimiters` — Return a list of all delimiters if valid.
    - `specials` — Return a list of all special characters.
    - `usable_scales` — Return a list of valid and enabled scales.
    - `usable_numerals` — Return a list of valid and enabled numerals.
    - `base_scale` — Return the base scale for the locale.

    #### Methods
    - `get_scale()` — Get the scale that contains a specific value in its terms list, if any.
    - `get_numeral()` — Get the numeral that contains a specific value in its terms list, if any.

    #### Raises
    - `InvalidLocaleError` when `spanish.json` does not exist or is malformed.
    - `InvalidParserError` when `self.parser` is not callable.
    - `NoValidScalesError` when no `Scale` is valid during validation.
    - `InvalidScaleError` when a `Scale` is invalid during initialization.
    - `InvalidNumeralError` when a `Numeral` is invalid during initialization.
    """

    def __init__(self, flags: FailureFlags | None = None, settings: ParserSettings | None = None) -> None:
        super().__init__("spanish.json", flags, settings)

    def __repr__(self) -> str:
        """Return a string representation of the locale with attributes included."""
        return f"Spanish(flags={repr(self.flags)}, settings={repr(self.settings)})"


class Guess:
    """---
    Represents an unknown `Locale` context used for parsing lengths of time.

    Does not contain any of the attributes of a `Locale`. Should never be used as a final `Locale`, but instead as an
    intermediary until a `Locale` is determined.

    #### Attributes
    - flags: `FailureFlags | None = None` — The flags that will cause parsing to fail.
        - If passed, flags loaded from the config will be overwritten.
    - settings: `ParserSettings | None = None` — The settings for the parser.
        - If passed, settings loaded from the config will be overwritten.
    - locales: `list[Locale]` — A list of all available `Locale` contexts.
        - Each `Locale` is initialized with `self`.

    #### Raises
    - `InvalidLocaleError` if any of the `Locale` configs are malformed or missing.
    """

    def __init__(
        self,
        flags: FailureFlags | None = None,
        settings: ParserSettings | None = None,
    ) -> None:
        self.flags: FailureFlags | None = flags
        self.settings: ParserSettings | None = settings
        self.locales: list[Locale] = [English(), Spanish()]

    def __str__(self) -> str:
        """Return the name of self."""
        return self.__class__.__name__

    def __repr__(self) -> str:
        """Return a string representation of the Guess with attributes included."""
        return f"Guess(flags={f'{repr(self.flags)}' if self.flags else None}, settings={repr(self.settings) if self.settings else None})"



from __future__ import annotations

from copy import deepcopy
from datetime import datetime, timedelta, timezone

from timelength.dataclasses import ParsedTimeLength, Scale
from timelength.errors import (
    InvalidScaleError,
    NotALocaleError,
    ParsedTimeDeltaError,
    PotentialDateTimeError,
    PotentialTimeDeltaError,
)
from timelength.locales import English, Guess, Locale


class TimeLength:
    """---
    Represents a length of time.

    #### Attributes
    - content: `str` — The string to be parsed.
    - locale: `Locale | Guess = English()` — The context used for parsing.
        - Will be set to the guessed locale if `Guess()` is passed.

    #### Properties
    - result: `ParsedTimeLength` — The result of parsing.

    #### Methods
    - `parse()` — Parse `self.content` based on `self.locale`.
        - Automatically called during initialization. Manually call if changes are made to `self.content` or `self.locale`.
    - `ago()` — Return a `datetime` from the past adjusted for `self.result`.
    - `hence()` — Return a `datetime` from the future adjusted for `self.result`.
    - `to_milliseconds()`, `to_minutes()`, `to_hours()`, `to_days()`, `to_weeks()`, `to_months()`,
        `to_years()`, `to_decades()`, `to_centuries()`
        - Convert the parsed duration to the respective units of each method.

    #### Raises
    - `NotALocaleError` when `locale` is not an instance of `Locale` or `Guess`.
    - `InvalidParserError` when the parser on `locale` is not callable.

    #### Example
    ```python
    from timelength import TimeLength

    tl = TimeLength("2 hours 30 minutes")
    print(f"Total Seconds: {tl.result.seconds}")
    ```
    """

    def __init__(self, content: str, locale: Locale | Guess = English()) -> None:
        self.content: str = str(content)

        # self.locale must always be a Locale, so if Guess is passed, temporarily set it to the first
        # locale in Guess().locales to prevent a wasted initialization while the best locale is found.
        self.locale: Locale = locale.locales[0] if isinstance(locale, Guess) else locale

        if not isinstance(self.locale, Locale):
            raise NotALocaleError(locale)

        self._result: ParsedTimeLength
        self.parse(guess_locale=locale if isinstance(locale, Guess) else False)

    @property
    def result(self) -> ParsedTimeLength:
        """The result of parsing."""
        return self._result

    def _parse_locale(self, locale: Locale) -> ParsedTimeLength:
        return locale.parser(self.content.strip(), locale)

    def parse(self, guess_locale: bool | Guess = False) -> None:
        """---
        Parse `self.content` using the parser attached to `self.locale`.

        #### Arguments
        - guess_locale: `bool | Guess = False` — Attempt each `Locale` and keep the best result.
            - The best result is the one with the least invalid results.
            - Pass an existing instance of `Guess` to prevent re-initializing every locale in `Guess().locales`.
            - Custom locales can be added to the pool by appending them to `Guess().locales` in the passed `Guess`.
            - Flags or settings passed to `Guess` will be used for each locale. If none are passed, the default for
                each locale will be used.
            - To use different flags or settings for each locale, access `Guess().locales` and set them individually.

        #### Updates
        - `self.result` with the outcome of parsing.
        - `self.locale` with the best locale if `guess_locale` is set.
        """

        if guess_locale is False:
            self._result = self._parse_locale(self.locale)
        else:
            guess: Guess = guess_locale if isinstance(guess_locale, Guess) else Guess()
            results: list[tuple[ParsedTimeLength, Locale]] = []

            for locale in guess.locales:
                locale.flags = guess.flags if guess.flags is not None else locale.flags
                locale.settings = guess.settings if guess.settings is not None else locale.settings

                results.append((self._parse_locale(locale), locale))

            # Sort most invalid to least invalid, breaking ties by least valid to most valid,
            # lastly breaking further ties by reverse alphabetical Locale name.
            results.sort(
                key=lambda res: (len(res[0].invalid), -len(res[0].valid), res[1].__class__.__name__), reverse=True
            )

            self._result, self.locale = results[-1]

    def ago(self, base: datetime = datetime.now(timezone.utc)) -> datetime:
        """---
        Get a `datetime` from the past.

        #### Arguments
        - base: `datetime = datetime.now(timezone.utc)` — The relative time to subtract from.

        #### Returns
        - A `datetime` representing `base` minus `self.result.delta`.

        #### Raises
        - `ParsedTimeDeltaError` when `self.result` exceeds the supported bounds of `timedelta`.
        - `PotentialDateTimeError` when the resultant `datetime` would exceed the supported bounds of `datetime`.
        """

        if self.result.delta is None:
            raise ParsedTimeDeltaError
        elif self._invalid_datetime(base, self.result.delta, subtract=True):
            raise PotentialDateTimeError
        else:
            return base - self.result.delta

    def hence(self, base: datetime = datetime.now(timezone.utc)) -> datetime:
        """---
        Get a `datetime` from the future.

        #### Arguments
        - base: `datetime = datetime.now(timezone.utc)` — The relative time to add to.

        #### Returns
        - A `datetime` representing `base` plus `self.result.delta`.

        #### Raises
        - `ParsedTimeDeltaError` when `self.result` exceeds the supported bounds of `timedelta`.
        - `PotentialDateTimeError` when the resultant `datetime` would exceed the supported bounds of `datetime`.
        """

        if self.result.delta is None:
            raise ParsedTimeDeltaError
        elif self._invalid_datetime(base, self.result.delta):
            raise PotentialDateTimeError
        else:
            return base + self.result.delta

    def to_milliseconds(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to milliseconds.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as milliseconds.

        #### Raises
        - `InvalidScaleError` when `self.locale.millisecond` is invalid.
        """

        return self._round(self.locale.millisecond, max_precision)

    def to_minutes(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to minutes.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as minutes.

        #### Raises
        - `InvalidScaleError` when `self.locale.minute` is invalid.
        """

        return self._round(self.locale.minute, max_precision)

    def to_hours(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to hours.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as hours.

        #### Raises
        - `InvalidScaleError` when `self.locale.hour` is invalid.
        """

        return self._round(self.locale.hour, max_precision)

    def to_days(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to days.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as days.

        #### Raises
        - `InvalidScaleError` when `self.locale.day` is invalid.
        """

        return self._round(self.locale.day, max_precision)

    def to_weeks(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to weeks.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as weeks.

        #### Raises
        - `InvalidScaleError` when `self.locale.week` is invalid.
        """

        return self._round(self.locale.week, max_precision)

    def to_months(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to months.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as months.

        #### Raises
        - `InvalidScaleError` when `self.locale.month` is invalid.
        """

        return self._round(self.locale.month, max_precision)

    def to_years(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to years.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as years.

        #### Raises
        - `InvalidScaleError` when `self.locale.year` is invalid.
        """

        return self._round(self.locale.year, max_precision)

    def to_decades(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to decades.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as decades.

        #### Raises
        - `InvalidScaleError` when `self.locale.decade` is invalid.
        """

        return self._round(self.locale.decade, max_precision)

    def to_centuries(self, max_precision: int | None = None) -> float:
        """---
        Return `self.result.seconds` converted to centuries.

        #### Arguments
        - max_precision: `int | None = None` — The maximum number of decimal places to include.

        #### Returns
        - A `float` representing `self.result.seconds` as centuries.

        #### Raises
        - `InvalidScaleError` when `self.locale.century` is invalid.
        """

        return self._round(self.locale.century, max_precision)

    def _round(self, scale: Scale, max_precision: int | None) -> float:
        if not scale.valid:
            raise InvalidScaleError(scale.singular)
        else:
            val = self.result.seconds / scale.scale
            return round(val, max_precision) if max_precision else val

    def _invalid_datetime(self, date: datetime, delta: timedelta, subtract: bool = False) -> bool:
        """Check if the resultant `datetime` would exceed the bounds supported by `datetime`."""

        date_sec = date.timestamp()
        delta_sec = delta.total_seconds()

        # datetime.min and datetime.max error on Windows when using .timestamp()
        # so it is necessary to manually create the datetime objects.
        min_datetime_timestamp = datetime(1, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc).timestamp()
        max_datetime_timestamp = datetime(9999, 12, 31, 23, 59, 59, 999999, tzinfo=timezone.utc).timestamp()

        return (
            (date_sec + delta_sec < min_datetime_timestamp or date_sec + delta_sec > max_datetime_timestamp)
            if not subtract
            else (date_sec - delta_sec < min_datetime_timestamp or date_sec - delta_sec > max_datetime_timestamp)
        )

    def _invalid_timedelta(self, first: timedelta, second: timedelta, subtract: bool = False) -> bool:
        """Check if the resultant `timedelta` would exceed the bounds supported by `timedelta`."""

        first_sec = first.total_seconds()
        second_sec = second.total_seconds()

        return (
            (
                first_sec + second_sec < timedelta.min.total_seconds()
                or first_sec + second_sec > timedelta.max.total_seconds()
            )
            if not subtract
            else (
                first_sec - second_sec < timedelta.min.total_seconds()
                or first_sec - second_sec > timedelta.max.total_seconds()
            )
        )

    def _convert_to_hhmmss(self, seconds: int | float) -> str:
        """Convert the passed seconds to `Days, HH:MM:SS.MS` format."""

        days, remainder = divmod(seconds, 86400)
        hours, remainder = divmod(remainder, 3600)
        minutes, remainder = divmod(remainder, 60)
        seconds, milliseconds = divmod(remainder, 1)
        decimal = round(milliseconds, 9)
        formatted = f"{int(hours):02}:{int(minutes):02}:{int(seconds):02}"

        if decimal:
            formatted += f".{str(decimal)[2:]}".rstrip("0").rstrip(".")

        if days > 0:
            return f"{int(days)} Day{'s' if days != 1 else ''}, {formatted}"
        else:
            return formatted

    def __str__(self) -> str:
        """Return `self.result.seconds` as a `Days, HH:MM:SS.MS` string."""
        return self._convert_to_hhmmss(self.result.seconds)

    def __repr__(self) -> str:
        """Return a string representation of the TimeLength with attributes included."""
        return f"TimeLength(content={repr(self.content)}, locale={repr(self.locale)})"

    def __add__(self, other: TimeLength | timedelta | float | int) -> TimeLength:
        """---
        Get the absolute sum of `self` and a `TimeLength`, a `timedelta`, or a number.

        #### Arguments
        - other: `TimeLength | timedelta | float | int` — The object to add to `self`.

        #### Returns
        - A `TimeLength` that represents the absolute sum of `self` and the passed value.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        base = self.locale.base_scale

        if isinstance(other, TimeLength):
            return TimeLength(
                content=f"{abs(self.result.seconds + other.result.seconds) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        elif isinstance(other, timedelta):
            return TimeLength(
                content=f"{abs(self.result.seconds + other.total_seconds()) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        elif isinstance(other, (float, int)):
            return TimeLength(
                content=f"{abs(self.result.seconds + other) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        else:
            return NotImplemented

    def __radd__(self, other: datetime | timedelta) -> datetime | timedelta:
        """---
        Get the sum of a `datetime` or a `timedelta` and `self`.

        #### Arguments
        - other: `datetime | timedelta` — The object to add `self` to.

        #### Returns
        - A `datetime` in the future by the amount of `self.result.delta`.
        - A `timedelta` that represents the sum of the passed value and `self.result.delta`.

        #### Raises
        - `PotentialDateTimeError` when the resultant `datetime` would exceed the supported bounds of `datetime`.
        - `ParsedTimeDeltaError` when `self.result` is a value exceeding the supported bounds of `timedelta`.
        - `PotentialTimeDeltaError` when the resultant `timedelta` would exceed the supported bounds of `timedelta`.
        """

        if isinstance(other, datetime):
            return self.hence(other)
        elif isinstance(other, timedelta):
            if self.result.delta is None:
                raise ParsedTimeDeltaError
            elif self._invalid_timedelta(other, self.result.delta):
                raise PotentialTimeDeltaError

            return self.result.delta + other
        else:
            return NotImplemented

    def __sub__(self, other: TimeLength | timedelta | float | int) -> TimeLength:
        """---
        Get the absolute difference between `self` and a `TimeLength`, a `timedelta`, or a number.

        #### Arguments
        - other: `TimeLength | timedelta | float | int` — The object to subtract from `self`.

        #### Returns
        - A `TimeLength` that represents the absolute difference between `self` and the passed value.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        base = self.locale.base_scale

        if isinstance(other, TimeLength):
            return TimeLength(
                content=f"{abs(self.result.seconds - other.result.seconds) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        elif isinstance(other, timedelta):
            return TimeLength(
                content=f"{abs(self.result.seconds - other.total_seconds()) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        elif isinstance(other, (float, int)):
            return TimeLength(
                content=f"{abs(self.result.seconds - other) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        else:
            return NotImplemented

    def __rsub__(self, other: datetime | timedelta) -> datetime | timedelta:
        """
        Get the difference between a a `datetime` or a `timedelta` and `self`.

        #### Arguments
        - other: `datetime | timedelta` — The object to subtract `self` from.

        #### Returns
        - A `datetime` in the past by the amount of `self.result.delta`.
        - A `timedelta` that represents the difference between the passed value and `self.result.delta`.

        #### Raises
        - `PotentialDateTimeError` when the resultant `datetime` would exceed the supported bounds of `datetime`.
        - `ParsedTimeDeltaError` when `self.result` is a value exceeding the supported bounds of `timedelta`.
        - `PotentialTimeDeltaError` when the resultant `timedelta` would exceed the supported bounds of `timedelta`.
        """

        if isinstance(other, datetime):
            return self.ago(other)
        elif isinstance(other, timedelta):
            if self.result.delta is None:
                raise ParsedTimeDeltaError
            elif self._invalid_timedelta(other, self.result.delta, True):
                raise PotentialTimeDeltaError("The resultant timedelta would exceed the supported bounds of timedelta.")

            return other - self.result.delta
        else:
            return NotImplemented

    def __mul__(self, other: float | int) -> TimeLength:
        """
        Get the absolute multiplication of `self` and a number.

        #### Arguments
        - other: `float | int` — The number to multiply `self` by.

        #### Returns
        - A `TimeLength` that represents the absolute multiplication of `self` and the passed number.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        if not isinstance(other, (float, int)):
            return NotImplemented

        base = self.locale.base_scale

        return TimeLength(
            content=f"{abs(self.result.seconds * other) / base.scale} {base.terms[0]}",
            locale=deepcopy(self.locale),
        )

    __rmul__ = __mul__

    def __truediv__(self, other: TimeLength | timedelta | float | int) -> TimeLength | float:
        """
        Get the division of `self` and a `TimeLength`, a `timedelta`, or a number.

        #### Arguments
        - other: `TimeLength | timedelta | float | int` — The object to divide `self` by.

        #### Returns
        - A `TimeLength` that represents the absolute division of `self` and the passed number.
        - A `float` that represents the division of `self` and the passed `TimeLength` or `timedelta`.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        if isinstance(other, TimeLength):
            return self.result.seconds / other.result.seconds
        elif isinstance(other, timedelta):
            return self.result.seconds / other.total_seconds()
        elif isinstance(other, (float, int)):
            base = self.locale.base_scale

            return TimeLength(
                content=f"{abs(self.result.seconds / other) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        else:
            return NotImplemented

    def __rtruediv__(self, other: timedelta) -> float:
        """
        Get the division of a `timedelta` and `self`.

        #### Arguments
        - other: `timedelta` — The object to divide by `self`.

        #### Returns
        - A `float` that represents the division of the passed `timedelta` and `self`.
        """

        if isinstance(other, timedelta):
            return other.total_seconds() / self.result.seconds
        else:
            return NotImplemented

    def __floordiv__(self, other: TimeLength | timedelta | float | int) -> TimeLength | float:
        """
        Get the floor division of `self` and a `TimeLength`, a `timedelta`, or a number.

        #### Arguments
        - other: `TimeLength | timedelta | float | int` — The object to divide `self` by.

        #### Returns
        - A `TimeLength` that represents the absolute floor division of `self` and the passed number.
        - A `float` that represents the floor division of `self` and the passed `TimeLength` or `timedelta`.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        if isinstance(other, TimeLength):
            return int(self.result.seconds // other.result.seconds)
        elif isinstance(other, timedelta):
            return int(self.result.seconds // other.total_seconds())
        elif isinstance(other, (float, int)):
            base = self.locale.base_scale

            return TimeLength(
                content=f"{abs(self.result.seconds // other) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        else:
            return NotImplemented

    def __rfloordiv__(self, other: timedelta) -> float:
        """
        Get the floor division of a `timedelta` and `self`.

        #### Arguments
        - other: `timedelta` — The object to divide by `self`.

        #### Returns
        - A `float` that represents the floor division of the passed `timedelta` and `self`.
        """

        if isinstance(other, timedelta):
            return int(other.total_seconds() // self.result.seconds)
        else:
            return NotImplemented

    def __mod__(self, other: TimeLength | timedelta) -> TimeLength:
        """
        Get the absolute modulo of `self` and a `TimeLength` or `timedelta`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to modulo `self` by.

        #### Returns
        - A `TimeLength` that represents the modulo of `self` and the passed `TimeLength` or `timedelta`.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        base = self.locale.base_scale

        if isinstance(other, TimeLength):
            return TimeLength(
                content=f"{abs(self.result.seconds % other.result.seconds) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        elif isinstance(other, timedelta):
            return TimeLength(
                content=f"{abs(self.result.seconds % other.total_seconds()) / base.scale} {base.terms[0]}",
                locale=deepcopy(self.locale),
            )
        else:
            return NotImplemented

    def __rmod__(self, other: timedelta) -> timedelta:
        """
        Get the modulo of a `TimeLength` or `timedelta` and `self`.

        #### Arguments
        - other: `timedelta` — The object to modulo by `self`.

        #### Returns
        - A timedelta that represents the modulo of the passed `timedelta` and `self`.

        #### Raises
        - `ParsedTimeDeltaError` when `self.result` is a value exceeding the supported bounds of `timedelta`.
        """

        if isinstance(other, timedelta):
            if self.result.delta is None:
                raise ParsedTimeDeltaError

            return other % self.result.delta
        else:
            return NotImplemented

    def __divmod__(self, other: TimeLength | timedelta) -> tuple[float, TimeLength]:
        """
        Get the divmod of `self` and a `TimeLength` or `timedelta`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to divmod `self` by.

        #### Returns
        - A tuple of a `float` and an absolute `TimeLength` that represent the divmod of `self` and the passed
            `TimeLength` or `timedelta`.
        """

        if isinstance(other, TimeLength):
            return (
                self.result.seconds // other.result.seconds,
                self % other,
            )
        elif isinstance(other, timedelta):
            return (
                self.result.seconds // other.total_seconds(),
                self % other,
            )
        else:
            return NotImplemented

    def __rdivmod__(self, other: timedelta) -> tuple[float, timedelta]:
        """
        Get the divmod of a `timedelta` and `self`.

        #### Arguments
        - other: `timedelta` — The object to divmod by `self`.

        #### Returns
        - A tuple of a `float` and a `timedelta` that represent the divmod of the passed `timedelta` and `self`.
        """

        if isinstance(other, timedelta):
            return (
                other.total_seconds() // self.result.seconds,
                other % self,
            )
        else:
            return NotImplemented

    def __pow__(self, other: float | int, mod: TimeLength | timedelta | None = None) -> TimeLength:
        """
        Get the absolute power of `self` and a number.

        #### Arguments
        - other: `float | int` — The number to raise `self` to.
        - mod: `TimeLength | timedelta = None` — The object to modulo the result by.

        #### Returns
        - A `TimeLength` that represents the absolute power of `self` and the passed number, optionally moduloed by mod.

        #### Raises
        - `NoValidScalesError` when no valid and enabled scales are found to perform the action.
        """

        if not isinstance(other, (float, int)):
            return NotImplemented
        elif mod is not None and not isinstance(mod, (TimeLength, timedelta)):
            return NotImplemented

        if not mod:
            result: float = abs(self.result.seconds**other)
        else:
            mod_seconds: float = mod.result.seconds if isinstance(mod, TimeLength) else mod.total_seconds()
            result: float = abs(self.result.seconds**other) % mod_seconds

        base = self.locale.base_scale

        return TimeLength(
            content=f"{result / base.scale} {base.terms[0]}",
            locale=deepcopy(self.locale),
        )

    def __bool__(self) -> bool:
        """Return if the parsing succeeded."""
        return self.result.success

    def __len__(self) -> int:
        """Return the length of `self.content`."""
        return len(self.content)

    def __abs__(self) -> TimeLength:
        """Return `self` unchanged as `TimeLength` is an absolute measurement."""
        return self

    def __pos__(self) -> TimeLength:
        """Return `self` unchanged as `TimeLength` is an absolute measurement."""
        return self

    def __neg__(self) -> TimeLength:
        """Return `self` unchanged as `TimeLength` is an absolute measurement."""
        return self

    def __invert__(self):
        return NotImplemented

    def __gt__(self, other: TimeLength | timedelta) -> bool:
        """
        Check if `self` is greater than `other`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is greater than `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return NotImplemented

        return self.result.seconds > (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())

    def __ge__(self, other: TimeLength | timedelta) -> bool:
        """
        Check if `self` is greater than or equal to other.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is greater than or equal to `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return NotImplemented

        return self.result.seconds >= (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())

    def __lt__(self, other: TimeLength | timedelta) -> bool:
        """
        Check if `self` is less than `other`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is less than `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return NotImplemented

        return self.result.seconds < (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())

    def __le__(self, other: TimeLength | timedelta) -> bool:
        """
        Check if `self` is less than or equal to `other`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is less than or equal to `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return NotImplemented

        return self.result.seconds <= (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())

    def __eq__(self, other: object) -> bool:
        """
        Check if `self` is equal to `other`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is equal to `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return False

        return self.result.seconds == (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())

    def __ne__(self, other: object) -> bool:
        """
        Check if `self` is not equal to `other`.

        #### Arguments
        - other: `TimeLength | timedelta` — The object to compare to.

        #### Returns
        - A `bool` indicating if `self` is not equal to `other` based on their seconds.
        """

        if not isinstance(other, (TimeLength, timedelta)):
            return True

        return self.result.seconds != (other.result.seconds if isinstance(other, TimeLength) else other.total_seconds())
